/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fLifting.h"
#include "fWake.h"
#include "wMshData.h"
#include "wGmshExport.h"
#include "wTag.h"
#include "wElement.h"
#include "wQuad4.h"
#include "wNode.h"
using namespace tbox;
using namespace fpm;

Lifting::Lifting(std::shared_ptr<MshData> _msh, std::vector<std::string> const &names, double xF) : Body(_msh, names[0])
{
    // Sanity check
    if (names.size() != 2)
    {
        std::stringstream err;
        err << "fpm::Lifting should be built with 2 physical groups but " << names.size() << " were given!\n";
        throw std::runtime_error(err.str());
    }

    // Check if wake is already available and create it otherwise
    std::cout << "Creating wake... " << std::flush;
    try
    {
        wake = new Wake(_msh, names[0] + "Wake", tag);
        std::cout << *wake << " already existing, nothing done." << std::endl;
    }
    catch (const std::out_of_range &)
    {
        // Create tags
        Tag *tagp = new Tag(static_cast<int>(msh->ptags.rbegin()->first + 1), tag->name + "Wake", 2);
        Tag *tage = new Tag(static_cast<int>(msh->etags.rbegin()->first + 1), tag->name + "Wake", 2);
        msh->ntags[tag->name + "Wake"] = tagp;
        msh->ptags[static_cast<int>(msh->ptags.rbegin()->first + 1)] = tagp;
        msh->etags[static_cast<int>(msh->etags.rbegin()->first + 1)] = tage;
        // Sort TE nodes (ascending along y-coordinate)
        Group te(_msh, names[1]);
        std::vector<Node *> teNodes;
        for (auto e : te.tag->elems)
            for (auto n : e->nodes)
                teNodes.push_back(n);
        std::sort(teNodes.begin(), teNodes.end(), [](Node *a, Node *b) -> bool { return a->pos(1) < b->pos(1); });
        teNodes.erase(std::unique(teNodes.begin(), teNodes.end()), teNodes.end());
        // Translate TE nodes (along x-coordinate)
        std::vector<Node *> wkNodes;
        for (auto n : teNodes)
        {
            Node *nN = new Node(msh->nodes.back()->no + 1, Eigen::Vector3d(xF, n->pos(1), n->pos(2)));
            wkNodes.push_back(nN);
            msh->nodes.push_back(nN);
        }
        // Create wake elements
        for (size_t i = 0; i < wkNodes.size() - 1; ++i)
        {
            std::vector<Node *> qnodes = {teNodes[i], wkNodes[i], wkNodes[i + 1], teNodes[i + 1]};
            msh->elems.push_back(new Quad4(msh->elems.back()->no + 1, tagp, tage, tag->elems[0]->parts, qnodes));
        }

        // Create wake and save modified mesh
        wake = new Wake(_msh, names[0] + "Wake", tag);
        std::cout << *wake << " created. " << std::flush;
        GmshExport writer(_msh);
        writer.save(_msh->name);
    }
}

Lifting::~Lifting()
{
    delete wake;
    std::cout << "~Lifting()\n";
}

void Lifting::write(std::ostream &out) const
{
    out << "fpm::Lifting " << *tag
        << " with " << *wake << std::endl;
}
